/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

/*_############################################################################
  _## 
  _##  tlv.cpp  
  _## 
  _##
  _##  
  _##########################################################################*/


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <tlv.h>


//------------------------------ Constructors -------------------------------

//------------------------------ Methods -------------------------------

int tlv::decodeTL(char *&buf, int &lengthOfBuffer, int &type, int &length)
{
  if (!buf || lengthOfBuffer <= 0)
    return 0;
  type = getInt(buf);
  length = getInt(buf);
  if (length < 0)
    return 0;
  lengthOfBuffer -= 2*sizeof(int);
  return 2*sizeof(int);
}

int tlv::decodeUINT(char *buf, int lengthOfBuffer, int T, unsigned int &V)
{
  int *p = (int *)&V;
  return decodeINT(buf, lengthOfBuffer, T, *p);
}

int tlv::decodeIpaddress(char *buf, int lengthOfBuffer, int T, ipaddress &V)
{
  int p;
  if (decodeINT(buf, lengthOfBuffer, T, p)) {
    V.setVal(p);
    return 1;
  }
  return 0;
}

int tlv::decodeINT(char *buf, int lengthOfBuffer, int type, int &V)
{
  int T, L;
  while (lengthOfBuffer > 0) {
    if (!tlv::decodeTL(buf, lengthOfBuffer, T, L)) {
      return 0;
    }
    if (T == type) {
      if (L != sizeof(int))
	return 0;
      V = getInt(buf);
      return 1;
    }
    buf += L;
    lengthOfBuffer -= L;
  }
  return 0;
}

int tlv::decodeFLOAT(char *buf, int lengthOfBuffer, int type, float &V)
{
  int T, L;
  while (lengthOfBuffer > 0) {
    if (!tlv::decodeTL(buf, lengthOfBuffer, T, L)) {
      return 0;
    }
    if (T == type) {
      if (L != sizeof(float))
	return 0;
      V = getFloat(buf);
      return 1;
    }
    buf += L;
    lengthOfBuffer -= L;
  }
  return 0;
}

int tlv:: decodeSTRING(char *buf, int lengthOfBuffer, int type, StringSock &V)
{
  int T, L;
  while (lengthOfBuffer > 0) {
    if (!tlv::decodeTL(buf, lengthOfBuffer, T, L)) {
      return 0;
    }
    if (T == type) {
      if (L) {
	if ((V.String = (char *)realloc(V.String, L)) == NULL) {
	  cout << "tlv.cpp::decodeSTRING:FATAL ERROR:: Error in malloc" << endl;
	  syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
	  exit(0);
	}
	memcpy(V.String, buf, L);
      }
      V.Length = L;
      return 1;
    }
    buf += L;
    lengthOfBuffer -= L;
  }
  return 0;
}

int tlv::encodeTL(char *&buf, int type, int length)
{
  if (!buf)
    return 0;
  putInt(buf, type);
  putInt(buf, length);
  return 2*sizeof(int);
}

int tlv::encodeBuffer(char *&buf, int type, int length, char *val)
{
  if (!encodeTL(buf, type, length))
    return 0;
  memcpy(buf, val, length);
  buf += length;
  return getLengthOfTLV(length);
}

/*
  buf & bufLength are modified after each of the T,L & V have been processed.
  The T nad L fields are stored in this class in local variables.
  The V field is only a pointer to the incoming buffer.
*/
int tlv::build(char *&buf, unsigned int &bufLength)
{
  if (State == TLV_COMPLETE)
    return State;
  if (State == TLV_TYPE_STATE) {
    if (bufLength < sizeof(int)) {
      return State;
    }
    Type = getInt(buf);
    bufLength -= sizeof(int);  // For type
    State = TLV_LEN_STATE;
    Length = 0;
  }
  if (State == TLV_LEN_STATE) {
    if (bufLength < sizeof(int)) {
      return State;
    }
    Length = getInt(buf);
    bufLength -= sizeof(int);  // For len
    State = TLV_VAL_STATE;
  }

  // Start pulling out the Val portion of the TLV
  if (Length > bufLength) {
    // Not enough data
    return State;
  }
  if (Length) {
    Value = (char *)realloc(Value, Length);
    if (Value) {
      memcpy(Value, buf, Length);
      buf += Length;
      bufLength -= Length;
    }
    else {
      cout << "tlv.cpp::build:FATAL ERROR:: Error in malloc" << endl;
      syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
      exit(0);
    }
  }
  State = TLV_COMPLETE;
  return State;
}
  
/*
  Extracts a full tlv out of the buffer & allocs space for a new tlv.
  If a tlv is passed as input, and the tlv is partially built, this routine
  will try and complete the tlv 
*/

int tlv::extract(tlv *&out, char *&buf, unsigned int &bufLen)
{
  if (!out) {
    out = new tlv();
  }
  // A new tlv is being built
  return out->build(buf, bufLen); // Return the state 
}
